/**
 * Hunter Adams (vha3@cornell.edu)
 * 
 * HARDWARE CONNECTIONS
 *  - GPIO 16 ---> VGA Hsync
 *  - GPIO 17 ---> VGA Vsync
 *  - GPIO 18 ---> 330 ohm resistor ---> VGA Green lo-bit |__ both wired to 150 ohm to ground 
 *  - GPIO 19 ---> 220 ohm resistor ---> VGA Green hi_bit |   and to VGA Green
 *  - GPIO 20 ---> 330 ohm resistor ---> VGA Blue
 *  - GPIO 21 ---> 330 ohm resistor ---> VGA Red
 *  - RP2040 GND ---> VGA GND
 *
 * RESOURCES USED
 *  - PIO state machines 0, 1, and 2 on PIO instance 0
 *  - DMA channels 0, 1, 2, and 3
 *  - 153.6 kBytes of RAM (for pixel color data)
 *
 * Protothreads v1.1.1
 * Threads:
 * core 0:
 * Graphics demo
 * blink LED25 
 * core 1:
 * Toggle gpio 4 
 * Serial i/o 
 * 
 * specifying a menu and menu item
 * Menu 
 * -- specify x,y,# items
 * -- menu base x,y
 * -- menu # of items
 * -- bounding box?
 * -- array of items
 * Item sruct
 * -- specify int/float, log/lin, initial value, title, color
 * -- item sequence number
 * -- item title (8 char)
 * -- item color (on black background)
 * -- value type int/float + linear/log
 * -- int value, min, max
 * -- float value, min, max
 * Appearance
 * -- Sequential text lines, each with own color
 * -- item title(8 char) -tab- item value(8 char) either %8d or %1.2e
 * -- selection should highlight something in the item, perhaps a small disc
 */
// ==========================================
// === VGA graphics library
// ==========================================
#include "vga16_graphics.h"
#include <stdio.h>
#include <stdlib.h>
// #include <math.h>
#include "pico/stdlib.h"
#include "hardware/pio.h"
#include "hardware/dma.h"
// // Our assembled programs:
// // Each gets the name <pio_filename.pio.h>
// #include "hsync.pio.h"
// #include "vsync.pio.h"
// #include "rgb.pio.h"

// ==========================================
// === protothreads globals
// ==========================================
#include "hardware/sync.h"
#include "hardware/timer.h"
#include "pico/multicore.h"
#include "string.h"
// protothreads header
#include "pt_cornell_rp2040_v1_1_1.h"

// colors
int col1, col2, col3 ;
// ==========================================

// Some globals for storing timer information
volatile unsigned int time_accum = 0;
unsigned int time_accum_old = 0 ;
char timetext[40];

// Timer interrupt
bool repeating_timer_callback(struct repeating_timer *t) {

    time_accum += 1 ;
    return true;
}

// ===========================================
// ADC setup 
// ===========================================
// setup ADC

short adc_x, adc_y ;

int ADC_setup(void){
  adc_init();
  adc_gpio_init(28);
  // p26 is ADC input 0 - gpio 28 is 2
  adc_select_input(2);
  //  48 MHz clock / 12.8 KHz sample rate
  //adc_set_clkdiv (3750);
  // 16 KHz  32 mSec window
  adc_set_clkdiv (3000);
  // free run
  adc_run(1);
  // result is in adc_hw->result
  // but we are going to forward that to the FIFO
  //adc_fifo_setup(bool en, bool dreq_en, uint16_t dreq_thresh, bool err_in_fifo, bool byte_shift)
  // fifo_enable, dreq_enable, thresh=1, no error, dont shift to 8-bits(for 8-bit PWM)
 // adc_fifo_setup(1,1,1,0,0);
  //
}

// ==================================================
// === set up a menu scheme for use with VGA
// ==================================================
// ALL items store float values
struct menu_item {
    // internal reference
    int item_index;
    // displayed string
    char item_name[12];
    // displayed color
    int item_color ;
    //1=float 0=int ;  
    int item_data_type ;
    // 1=log 0=linear
    int item_inc_type ;
    // increment delta if linear, ratio if log
    float item_increment ;
    // the current value, min, max 
    // al updates are performed on float, then copied to int
    int item_int_value ;
    float item_float_value ;
    float item_float_min ;
    float item_float_max ;
};
// ======
// build the menu array of items
int menu_x = 400, menu_y = 200 ;
#define menu_num_items 4
struct menu_item menu[menu_num_items] ;
// =====
void draw_menu(void){
    for(int i=0; i<menu_num_items; i++){
        // display the name
        setTextColor(menu[i].item_color) ;
        setCursor(menu_x, menu_y+i*10) ;
        setTextSize(1) ;
        writeString(menu[i].item_name) ;
        // display the value
        setTextColor(WHITE) ;
        setCursor(menu_x+12*6, menu_y+i*10) ;
        char value_str[15] ;
        // get the data type in bit 0
        if (menu[i].item_data_type == 0) 
            sprintf(value_str, "%d", (int)menu[i].item_float_value) ;
        if (menu[i].item_data_type == 1) 
            sprintf(value_str, "%1.2e", menu[i].item_float_value);
        writeString(value_str) ;
    }
}
// =====
// direction is 1 for increase,  -1 for decrease
void change_value(int index, int direction){
    for(int i=0; i<menu_num_items; i++){

    }
}

// ==================================================
// === graphics demo -- RUNNING on core 0
// ==================================================
static PT_THREAD (protothread_graphics(struct pt *pt)) {
    PT_BEGIN(pt);
    // the protothreads interval timer
    PT_INTERVAL_INIT() ;

    // test menu setup
    // item identifier
    menu[0].item_index = 1 ;
    menu[1].item_index = 2 ;
    menu[2].item_index = 3 ;
    menu[3].item_index = 4 ;
    // color
    menu[0].item_color = GREEN ;
    menu[1].item_color = GREEN ;
    menu[2].item_color = RED ;
    menu[3].item_color = RED ;
    //
    sprintf(menu[0].item_name, "Int, lin ") ;
    sprintf(menu[1].item_name, "Int, log ") ;
    sprintf(menu[2].item_name, "Float, lin ") ;
    sprintf(menu[3].item_name, "Float, log ") ;
    //
    menu[0].item_data_type = 0 ;// int
    menu[1].item_data_type = 0 ;// int
    menu[2].item_data_type = 1 ;// float
    menu[3].item_data_type = 1 ;// float
    //
    menu[0].item_inc_type = 0 ;// lin
    menu[1].item_inc_type = 1 ;// log
    menu[2].item_inc_type = 0 ;// lin
    menu[3].item_inc_type = 1 ;// log
    // 
    menu[0].item_float_value = 12 ;// 
    menu[1].item_float_value = 3000 ;// 
    menu[2].item_float_value = 1.5 ;// 
    menu[3].item_float_value = -1000 ;// 
   // 
   draw_menu();

    // Draw some filled rectangles at top of screen
    fillRect(64, 0, 176, 50, BLUE); // blue box
    fillRect(250, 0, 176, 50, DARK_ORANGE); // red box
    fillRect(435, 0, 176, 50, GREEN); // green box

    // Write some intro text
    setTextColor(WHITE) ;
    setCursor(65, 0) ;
    setTextSize(1) ;
    writeString("Raspberry Pi Pico") ;
    setCursor(65, 10) ;
    writeString("Graphics interact demo") ;
    setCursor(65, 20) ;
    writeString("Hunter Adams") ;
    setCursor(65, 30) ;
    writeString("vha3@cornell.edu") ;
    setCursor(445, 10) ;
    setTextColor(BLACK) ;
    setTextSize(1) ;
    writeString("Protothreads rp2040 v1.1.1") ;
    setCursor(445, 20) ;
    writeString("Mod for 16 colors (brl4)") ;
    setTextColor(WHITE) ;

    // Setup a 1Hz timer
    static struct repeating_timer timer;
    add_repeating_timer_ms(-1000, repeating_timer_callback, NULL, &timer);

    // draw color map blocks
    // and name them
     char video_buffer[32];
    setTextColor2(WHITE, BLACK) ;
    setTextSize(1) ;
    for (int i=0; i<4; i++) {
      for (int j=0; j<4; j++){
        fillRect(i*70+20, 150+j*70, 60, 60, i+4*j);  
        setCursor(i*70+20, 150+j*70) ; 
        sprintf(video_buffer, "%2d", i+4*j);
        writeString(video_buffer) ;
      }
    }
    // first row of colors
    setCursor(0*70+20, 200+0*70) ; 
    writeString("BLACK") ;
    setCursor(1*70+20, 200+0*70) ; 
    writeString("DARK_GREEN") ;
    setCursor(2*70+20, 200+0*70) ; 
    writeString("MED_GREEN") ;
    setCursor(3*70+20, 200+0*70) ; 
    writeString("GREEN") ;
    // second row of colors
    setCursor(0*70+20, 200+1*70) ; 
    writeString("DARK_BLUE") ;
    setCursor(1*70+20, 200+1*70) ; 
    writeString("BLUE") ;
    setCursor(2*70+20, 200+1*70) ; 
    writeString("LIGHT_BLUE") ;
    setCursor(3*70+20, 200+1*70) ; 
    writeString("CYAN") ;
    // thrid row of colors
    setCursor(0*70+20, 200+2*70) ; 
    writeString("RED") ;
    setCursor(1*70+20, 200+2*70) ; 
    writeString("DARK_ORANGE") ;
    setCursor(2*70+20, 200+2*70) ; 
    writeString("ORANGE") ;
    setCursor(3*70+20, 200+2*70) ; 
    writeString("YELLOW") ;
    // fourth row of colors
    setCursor(0*70+20, 200+3*70) ; 
    writeString("MAGENTA") ;
    setCursor(1*70+20, 200+3*70) ; 
    writeString("PINK") ;
    setCursor(2*70+20, 200+3*70) ; 
    writeString("LIGHT_PINK") ;
    setCursor(3*70+20, 200+3*70) ; 
    writeString("WHITE") ;
    //
       

    while(true) {

        // Modify the color chooser
        //if (color_index ++ == 15) color_index = 0 ;
        //color_index = (rand() & 0x0f) ;
        // A row of filled circles
        //fillRect(disc_x, 100, 30, 30, color_index);
        ///disc_x += 32 ;
        //if (disc_x > 600) disc_x = 32;
       
        // A brief nap
        PT_YIELD_usec(500000) ;
   }
   PT_END(pt);
} // graphics thread

// ==================================================
// === toggle25 thread on core 0
// ==================================================
// the on-board LED blinks
static PT_THREAD (protothread_toggle25(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    
     // set up LED p25 to blink
     gpio_init(25) ;	
     gpio_set_dir(25, GPIO_OUT) ;
     gpio_put(25, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        // yield time 0.1 second
        //PT_YIELD_usec(100000) ;
        PT_YIELD_INTERVAL(100000) ;

        // toggle the LED on PICO
        LED_state = LED_state? false : true ;
        gpio_put(25, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread


// ==================================================
// === toggle gpio 4 thread -- RUNNING on core 1
// ==================================================
// toggle gpio 4 
static PT_THREAD (protothread_toggle_gpio4(struct pt *pt))
{
    PT_BEGIN(pt);
    static bool LED_state = false ;
    //
     // set up LED gpio 4 to blink
     gpio_init(4) ;	
     gpio_set_dir(4, GPIO_OUT) ;
     gpio_put(4, true);
     // data structure for interval timer
     PT_INTERVAL_INIT() ;

      while(1) {
        //
        PT_YIELD_INTERVAL(20) ;
        // toggle gpio 4
        LED_state = !LED_state ;
        gpio_put(4, LED_state);
        //
        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // blink thread

// ==================================================
// === user's serial input thread on core 1
// ==================================================
// serial_read an serial_write do not block any thread
// except this one
static PT_THREAD (protothread_serial(struct pt *pt))
{
    PT_BEGIN(pt);
      static int test_in1, test_in2, sum ;
      //
      while(1) {
        // print prompt
        sprintf(pt_serial_out_buffer, "input color1, color2, color3: ");
        // spawn a thread to do the non-blocking write
        serial_write ;

        // spawn a thread to do the non-blocking serial read
         serial_read ;
         printf("returned");
        // convert input string to number
        sscanf(pt_serial_in_buffer,"%d %d %d",  &col1, &col2, &col3) ;
        fillRect(400, 300, 60, 60, col1); 
        fillRect(460, 300, 60, 60, col2); 
        fillRect(520, 300, 60, 60, col3); 

        // NEVER exit while
      } // END WHILE(1)
  PT_END(pt);
} // serial thread

// ========================================
// === core 1 main -- started in main below
// ========================================
void core1_main(){ 
  //
  //  === add threads  ====================
  // for core 1
  pt_add_thread(protothread_toggle_gpio4) ;
  pt_add_thread(protothread_serial) ;
  //
  // === initalize the scheduler ==========
  pt_schedule_start ;
  // NEVER exits
  // ======================================
}

// ========================================
// === core 0 main
// ========================================
int main(){
  // set the clock
  //set_sys_clock_khz(250000, true); // 171us
  // start the serial i/o
  stdio_init_all() ;
  // announce the threader version on system reset
  printf("\n\rProtothreads RP2040 v1.11 two-core\n\r");

  // Initialize the VGA screen
  initVGA() ;
     
  // start core 1 threads
  multicore_reset_core1();
  multicore_launch_core1(&core1_main);

  // === config threads ========================
  // for core 0
  pt_add_thread(protothread_graphics);
  pt_add_thread(protothread_toggle25);
  //
  // === initalize the scheduler ===============
  pt_schedule_start ;
  // NEVER exits
  // ===========================================
} // end main